import random
import math

import pygame

from pygame import draw as pydraw

from pygame import key as pykey

from pygame.math import Vector2 as v2
from pygame.math import Vector3 as v3
from pygame.locals import *

GOAL_RANGE = 1/7

class CharacterData:
    in_place = None
    left = None
    right = None
    back = None

    def __init__(self, in_place, left, right, back):
        self.in_place = in_place
        self.left = left
        self.right = right
        self.back = back

def frustum(x, y, z, w, h):
    # fx = (x-w/2) * (1 - 0.6 * ((h-y)/h)) + w/2
    # fy = h/2 + math.log(y+1)/math.log(h+1)*(y*2*0.3) - h/4
    # fy = math.atan(y)*h/2
    # fx = x
    # fy = y
    x -= w/2
    y -= h/2
    x *= 1.2
    y *= 1.2
    z += 880
    K = 640
    y -= K
    if (-y+K == 0):
        return 0, 0
    fx = K*x/(-y+K)
    fy = K*z/(-y+K)

    fx += w/2
    fy -= 10

    return fx, fy

def frustum_scale(y, xs, zs, w, h):
    # s = 1 + (y/h*2*0.3)
    y -= h/2
    y *= 1.1
    K = 640
    y -= K
    if (-y+K == 0):
        return 0, 0
    xs = K*xs/(-y+K)
    ys = K*zs/(-y+K)

    return 2*xs, 2*ys

class EventType:
    KEYDOWN = 0,
    KEYUP   = 1

class BallPower:
    NONE   = 0,
    FREEZE = 1

class GameObject:
    pos = v2()
    vel = v2()
    acc = v2()
    on_ground = True
    alive = True
    z = 0.0
    vel_z = 0.0
    gravity = -0.5

    def think(self, game, w, h):
        pass
    def draw(self, screen, w, h):
        pass

class FreezePowerup(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 49
        self.h = 49
        self.frames = []
        img = pygame.image.load("freeze_powerup.png")
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)
        self.powerup_sound = pygame.mixer.Sound('freeze_powerup.wav')

    def think(self, game, w, h):
        for obj in game.objects:
            if not isinstance(obj, Player) and not isinstance(obj, Ball):
                continue
            if obj.z > 0:
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.powerup_sound.play()
                self.alive = False
                # Give all balls the power
                for obj in game.objects:
                    if isinstance(obj, Ball):
                        obj.power = BallPower.FREEZE

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # pydraw.rect(screen, (255, 255, 0), (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class SwapPowerup(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 49
        self.h = 49
        self.frames = []
        img = pygame.image.load("swap_powerup.png")
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)
        self.powerup_sound = pygame.mixer.Sound('swap_powerup.wav')

    def think(self, game, w, h):
        if not self.alive:
            return
        for obj in game.objects:
            if not isinstance(obj, Player) and not isinstance(obj, Ball):
                continue
            if obj.z > 0:
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.powerup_sound.play()
                self.alive = False
                for obj2 in game.objects:
                    if not isinstance(obj2, Player):
                        continue
                    if obj2 is obj:
                        continue
                    tx         = obj2.pos.x
                    ty         = obj2.pos.y
                    tz         = obj2.z

                    obj2.pos.x = obj.pos.x
                    obj2.pos.y = obj.pos.y
                    obj2.z     = obj.z

                    obj.pos.x  = tx
                    obj.pos.y  = ty
                    obj.z      = tz

                    obj.vel_z, obj2.vel_z = obj2.vel_z, obj.vel_z
                    return

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # pydraw.rect(screen, (255, 255, 0), (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class ThrowPowerup(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 49
        self.h = 49
        self.frames = []
        img = pygame.image.load("throw_powerup.png")
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)
        self.powerup_sound = pygame.mixer.Sound('throw_powerup.wav')

    def think(self, game, w, h):
        if not self.alive:
            return
        for obj in game.objects:
            if not isinstance(obj, Player) and not isinstance(obj, Ball):
                continue
            if obj.z > 0:
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.powerup_sound.play()
                self.alive = False
                for obj2 in game.objects:
                    if not isinstance(obj2, Player):
                        continue
                    if obj2 is obj:
                        continue
                    obj2.vel_z = 50
                break

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # pydraw.rect(screen, (255, 255, 0), (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class ConfusePowerup(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 49
        self.h = 49
        self.frames = []
        img = pygame.image.load("confuse_powerup.png")
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)
        self.powerup_sound = pygame.mixer.Sound('confuse_powerup.wav')

    def think(self, game, w, h):
        if not self.alive:
            return
        for obj in game.objects:
            if not isinstance(obj, Player) and not isinstance(obj, Ball):
                continue
            if obj.z > 0:
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.powerup_sound.play()
                self.alive = False
                for obj2 in game.objects:
                    if not isinstance(obj2, Player):
                        continue
                    if obj2 is obj:
                        continue
                    obj2.confused = not obj2.confused
                    obj2.confuse_start = pygame.time.get_ticks()
                break

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # pydraw.rect(screen, (255, 255, 0), (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class Meteor(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 100*10
        self.h = 100*10
        self.z = 1000
        self.meteor_sound = pygame.mixer.Sound('meteor.wav')
        self.frames = []
        self.start = pygame.time.get_ticks()
        img = pygame.image.load("meteor.png")
        for i in range(35):
            sub = img.subsurface((0, i*100, 100, 100))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)

    def think(self, game, w, h):
        if not self.alive:
            return
        self.z -= 10
        if pygame.time.get_ticks() > self.start + 3300:
            self.alive = False
            game.objects.append(Firestorm(v2(self.pos.x, self.pos.y)))

    def draw(self, screen, w, h):
        if not self.alive:
            return
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[((pygame.time.get_ticks()-self.start) // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class Firestorm (GameObject):
    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 200*10
        self.h = 50*10
        # self.meteor_sound = pygame.mixer.Sound('firestorm.wav')
        self.frames = []
        self.alive = True
        self.start = pygame.time.get_ticks()
        img = pygame.image.load("firestorm.png")
        for i in range(26):
            sub = img.subsurface((0, i*50, 200, 50))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)

    def think(self, game, w, h):
        if not self.alive:
            return
        if pygame.time.get_ticks() > self.start + 2460/10 * 3:
            self.alive = False
            for obj in game.objects:
                if isinstance(obj, MeteorPowerup) or isinstance(obj, FreezePowerup) or isinstance(obj, ThrowPowerup) or isinstance(obj, SwapPowerup) or isinstance(obj, ConfusePowerup):
                    obj.alive = False
                    continue
                if not isinstance(obj, Player):
                    continue
                obj.vel = obj.pos * 10000
                obj.stunned = True
                obj.stun_start = pygame.time.get_ticks()

    def draw(self, screen, w, h):
        if not self.alive:
            return
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[((pygame.time.get_ticks()-self.start) // 30) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class MeteorPowerup(GameObject):

    def __init__(self, pos: v2):
        self.pos = pos
        self.w = 49
        self.h = 49
        self.frames = []
        img = pygame.image.load("meteor_powerup.png")
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)
        self.powerup_sound = pygame.mixer.Sound('meteor_powerup.wav')

    def think(self, game, w, h):
        if not self.alive:
            return
        for obj in game.objects:
            if not isinstance(obj, Player) and not isinstance(obj, Ball):
                continue
            if obj.z > 0:
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.powerup_sound.play()
                self.alive = False
                game.objects.append(Meteor(v2(0, 0)))
                break

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # pydraw.rect(screen, (255, 255, 0), (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class Spring(GameObject):

    def __init__(self, pos: v2, dir: v2):
        self.pos = pos
        self.dir = dir
        self.w = 32
        self.h = 32
        self.spring_sound = pygame.mixer.Sound('spring.wav')
        self.last_hit = None
        img = pygame.image.load("spring.png")
        self.frames = []
        for i in range(6):
            sub = img.subsurface((0, i*7, 8, 7))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)

    def think(self, game, w, h):
        for obj in game.objects:
            if not isinstance(obj, Ball):
                continue
            t = obj.pos - self.pos
            l = t.length()
            if 0 <= l <= self.w/2 + obj.w/2:
                self.last_hit = pygame.time.get_ticks()
                self.spring_sound.play()
                obj.vel.x = self.dir.x
                obj.vel.y = self.dir.y

    def draw(self, screen, w, h):
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        if self.last_hit:
            scaled = pygame.transform.scale(self.frames[((pygame.time.get_ticks()-self.last_hit) // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
            if self.last_hit + 600 < pygame.time.get_ticks():
                self.last_hit = None
        else:
            scaled = pygame.transform.scale(self.frames[0], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class Wavepoint:

    def __init__(self, pos, vel):
        self.pos = pos
        self.vel = vel
        self.alive = True

    def think(self, w, h):
        if not self.alive:
            return
        self.pos += self.vel
        if self.pos.x < -w/2:
            self.pos.x = -w/2
            self.alive = False
        if self.pos.y < -h/2:
            self.pos.y = -h/2
            self.alive = False
        if self.pos.x > w/2:
            self.pos.x = w/2
            self.alive = False
        if self.pos.y > h/2:
            self.pos.y = h/2
            self.alive = False

class Wave:

    def __init__(self, pos, lifetime, speed):
        self.points = []
        self.alive = True
        self.pos = pos
        self.range = 0
        self.lifetime = lifetime
        self.speed = speed
        for i in range(60):
            x = math.cos(i/60 * 2*math.pi)
            y = math.sin(i/60 * 2*math.pi)
            pos = v2(pos.x, pos.y)
            vel = v2(x, y) * self.speed
            self.points.append(Wavepoint(pos, vel))

    def think(self, game, w, h):
        self.lifetime -= 1
        if self.lifetime <= 0:
            self.alive = False
            return
        self.range += self.speed
        for p in self.points:
            p.think(w, h)

    def draw(self, screen, w, h):
        if not self.alive:
            return
        last_x, last_y = None, None
        for p in self.points + [self.points[0]]:
            if last_x and last_y:
                x1, y1 = frustum( last_x+w/2,  last_y+h/2,    0, w, h)
                x2, y2 = frustum(p.pos.x+w/2, p.pos.y+h/2,    0, w, h)
                # x3, y3 = frustum(p.pos.x+w/2, p.pos.y+h/2, -100, w, h)
                # x4, y4 = frustum( last_x+w/2,  last_y+h/2, -100, w, h)
                # pydraw.polygon(screen, (0, 0, 255), [(x1, y1), (x2, y2), (x3, y3), (x4, y4)])
                pydraw.line(screen, (0, 0, 255), (x1, y1), (x2, y2))
            last_x = p.pos.x
            last_y = p.pos.y

class Ball(GameObject):
    w = 40
    h = 38

    def __init__(self):
        self.pos = v2()
        self.pos.x = 0
        self.pos.y = 0
        self.z = 100
        self.friction = 0.03

        img = pygame.image.load("freezeball.png")
        self.freeze_frames = []
        for i in range(6):
            sub = img.subsurface((0, i*19, 20, 19))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.freeze_frames.append(sub)

        img = pygame.image.load("ball.png")
        self.frames = []
        for i in range(17):
            sub = img.subsurface((0, i*19, 20, 19))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.frames.append(sub)

        self.hit_sound = pygame.mixer.Sound('hit.wav')
        self.goal_sound = pygame.mixer.Sound("goal.wav")
        self.freeze_sound = pygame.mixer.Sound('player_freeze.wav')

        self.power = BallPower.NONE

    def apply_physics(self, game, w, h):
        self.z += self.vel_z
        self.vel_z += self.gravity
        if self.z > 0:
            self.on_ground = False
        elif self.z < 0:
            self.on_ground = True
            self.z = 0.0
            self.vel_z = 0.0

        self.vel *= 1 - self.friction
        self.vel += self.acc
        self.pos += self.vel
        self.acc.x = 0
        self.acc.y = 0

        # edges
        if self.pos.x < -w/2 + self.w/2:
            # Are we inside a goal?
            if -GOAL_RANGE*h < self.pos.y < GOAL_RANGE*h:
                self.goal_sound.play()
                game.scores[1] += 1
                self.pos.x = 0
                self.pos.y = 0
                self.vel.x = 0
                self.vel.y = 0
                self.z = 10000
            else:
                self.pos.x = -w/2 + self.w/2
                self.vel.x = -self.vel.x
        if self.pos.y < -h/2:
            self.pos.y = -h/2
            self.vel.y = -self.vel.y
        if self.pos.x > w/2 - self.w/2:
            # Are we inside a goal?
            if -GOAL_RANGE*h < self.pos.y < GOAL_RANGE*h:
                self.goal_sound.play()
                game.scores[0] += 1
                self.pos.x = 0
                self.pos.y = 0
                self.vel.x = 0
                self.vel.y = 0
                self.z = 10000
            else:
                self.pos.x = w/2 - self.w/2
                self.vel.x = -self.vel.x
        if self.pos.y > h/2:
            self.pos.y = h/2
            self.vel.y = -self.vel.y

    def think(self, game, w, h):
        # for obj in game.objects:
        #     if obj.on_ground and self.on_ground and math.fabs(obj.pos.x - self.pos.x) <= self.w and math.fabs(obj.pos.y - self.pos.y) <= self.h:
        #         t = (self.pos - obj.pos)
        #         if t.length() > 0:
        #             self.hit_sound.play()
        #             self.vel = t.normalize() * 20
        #             self.vel_z = 10
        if self.power == BallPower.FREEZE:
            for obj in game.objects:
                if isinstance(obj, Player):
                    if obj.z > 0:
                        continue
                    t = (self.pos - obj.pos)
                    l = t.length()
                    if 0 <= l <= self.w/2 + obj.w / 2:
                        obj.stunned = True
                        obj.stun_start = pygame.time.get_ticks()
                        self.freeze_sound.play()
                        self.power = BallPower.NONE
                        break

        if self.on_ground:
            for wave in game.waves:
                t = (self.pos - wave.pos)
                l = t.length()
                if l > 0 and wave.range - self.w/2 <= l <= wave.range + self.w/2:
                    self.hit_sound.play()
                    self.vel = t.normalize() * wave.speed
                    self.vel_z = 10
                    break
        self.apply_physics(game, w, h)

    def draw(self, screen, w, h):
        # Shadow
        shadow_scale = max(0.3, (150.0 - self.z) / 150.0)
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w*0.75*shadow_scale, self.h*0.75/2*shadow_scale, w, h)
        pydraw.ellipse(screen, (0, 0, 0), (int(x-scaled_x/2), int(y-scaled_y/2), int(scaled_x), int(scaled_y)))
        # Ball
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # Freeze powerup
        if self.power == BallPower.FREEZE:
            x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
            scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
            scaled = pygame.transform.scale(self.freeze_frames[(pygame.time.get_ticks() // 100) % len(self.freeze_frames)], (int(scaled_x), int(scaled_y)))
            screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

class Player(GameObject):
    left_pressed = False
    up_pressed = False
    down_pressed = False
    right_pressed = False
    charge_start = None
    timer = pygame.time.Clock()

    def __init__(self, button_up, button_left, button_down, button_right, button_jump, game, char_data):
        self.button_up = button_up
        self.button_left = button_left
        self.button_down = button_down
        self.button_right = button_right
        self.button_jump = button_jump
        self.pos = v2()
        self.vel = v2()
        self.acc = v2()
        self.pounding = False
        self.charging = False
        self.stunned = False
        self.confused = False
        self.stun_start = None
        self.overheat = 0
        self.w = 64
        self.h = 64

        self.exploding = False
        self.explosion_counter = 0

        self.jump_sound = pygame.mixer.Sound('jump.wav')
        self.pound_sound = pygame.mixer.Sound('pound.wav')

        img = pygame.image.load(char_data.in_place[0])
        self.in_place_frames = []
        for i in range(char_data.in_place[1]):
            sub = img.subsurface((0, i*char_data.in_place[3], char_data.in_place[2], char_data.in_place[3]))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.in_place_frames.append(sub)

        img = pygame.image.load(char_data.left[0])
        self.left_frames = []
        for i in range(char_data.left[1]):
            sub = img.subsurface((0, i*char_data.left[3], char_data.left[2], char_data.left[3]))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.left_frames.append(sub)

        img = pygame.image.load(char_data.right[0])
        self.right_frames = []
        for i in range(char_data.right[1]):
            sub = img.subsurface((0, i*char_data.right[3], char_data.right[2], char_data.right[3]))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.right_frames.append(sub)

        img = pygame.image.load(char_data.back[0])
        self.back_frames = []
        for i in range(char_data.back[1]):
            sub = img.subsurface((0, i*char_data.back[3], char_data.back[2], char_data.back[3]))
            sub = pygame.transform.scale(sub, (self.w, self.h))
            self.back_frames.append(sub)

        self.frames = self.in_place_frames

        img = pygame.image.load("explosion.png")
        self.explosion_frames = []
        for i in range(36):
            sub = img.subsurface((0, i*32, 96, 32))
            sub = pygame.transform.scale(sub, (96, 32))
            self.explosion_frames.append(sub)

        img = pygame.image.load("stun_status_effect.png")
        self.stun_frames = []
        for i in range(4):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (7, 7))
            self.stun_frames.append(sub)

        img = pygame.image.load("confuse_status_effect.png")
        self.confuse_frames = []
        for i in range(8):
            sub = img.subsurface((0, i*7, 7, 7))
            sub = pygame.transform.scale(sub, (7, 7))
            self.confuse_frames.append(sub)

    def apply_physics(self, game, w, h):
        self.z += self.vel_z
        self.vel_z += self.gravity
        if self.z > 0:
            self.on_ground = False
        elif self.z < 0:
            if not self.on_ground:
                if self.pounding and self.pound_sound:
                    self.pound_sound.play()
                    self.exploding = True
                    self.explosion_counter = pygame.time.get_ticks()
                    for i in range(50):
                        x = math.cos(i/50 * 2*math.pi)
                        y = math.sin(i/50 * 2*math.pi)
                        pos = v3(self.pos.x, self.pos.y, 0)
                        vel = v3(x, y, 0.5*random.random()) * 4;
                        game.spawn_particle(Particle(60, pos, vel, v3(0, 0, -0.1)))
                    game.waves.append(Wave(v2(self.pos), 30, 8))
                self.pounding = False
            self.on_ground = True
            self.z = 0.0
            self.vel_z = 0.0

        self.vel *= 1 - self.friction
        self.vel += self.acc
        self.pos += self.vel
        self.acc.x = 0
        self.acc.y = 0

        # edges
        if self.pos.x < -w/2:
            self.pos.x = -w/2
            self.vel.x = 0
        if self.pos.y < -h/2:
            self.pos.y = -h/2
            self.vel.y = 0
        if self.pos.x > w/2:
            self.pos.x = w/2
            self.vel.x = 0
        if self.pos.y > h/2:
            self.pos.y = h/2
            self.vel.y = 0

    def think(self, game, w, h):
        speed = 1.6
        max_speed = 3.0
        if self.overheat > 0:
            self.overheat -= 1
        if self.charging:
            speed = 0.3
            max_speed = 0.8
        if self.confused and self.confuse_start + 5000 < pygame.time.get_ticks():
            self.confused = False
        if self.stunned and self.stun_start + 2000 < pygame.time.get_ticks():
            self.stunned = False
        if not self.stunned:
            if self.confused and self.left_pressed or not self.confused and self.right_pressed:
                self.acc.x += 1.0 * speed
            if self.confused and self.right_pressed or not self.confused and self.left_pressed:
                self.acc.x -= 1.0 * speed
            if self.confused and self.up_pressed or not self.confused and self.down_pressed:
                self.acc.y += 1.0 * speed
            if self.confused and self.down_pressed or not self.confused and self.up_pressed:
                self.acc.y -= 1.0 * speed
            if (self.up_pressed != self.down_pressed) or (self.left_pressed != self.right_pressed):
                self.friction = 0.0
            else:
                self.friction = 0.1

            threshold = 1.4
            if self.vel.x > threshold:
                self.frames = self.right_frames
            elif self.vel.x < -threshold:
                self.frames = self.left_frames
            elif self.vel.y < -threshold:
                self.frames = self.back_frames
            elif self.vel.y > threshold:
                self.frames = self.in_place_frames
            else:
                self.frames = self.in_place_frames
        else:
            self.friction = 1.0

        self.apply_physics(game, w, h)
        if self.vel.length() > max_speed:
            self.vel.normalize_ip()
            self.vel *= max_speed

    def draw(self, screen, w, h):
        # Shadow
        shadow_scale = max(0.3, (200.0 - self.z) / 200.0)
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, 0, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w*0.65*shadow_scale, self.h*0.5*0.65*shadow_scale, w, h)
        pydraw.ellipse(screen, (0, 0, 0), (int(x-scaled_x/2), int(y-scaled_y/2), int(scaled_x), int(scaled_y)))
        # Image
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
        scaled_x, scaled_y = frustum_scale(self.pos.y, self.w, self.h, w, h)
        scaled = pygame.transform.scale(self.frames[(pygame.time.get_ticks() // 100) % len(self.frames)], (int(scaled_x), int(scaled_y)))
        screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))
        # Explosion
        if self.exploding:
            if self.explosion_counter + 36/100 * 1000 < pygame.time.get_ticks():
                self.exploding = False
            x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z, w, h)
            scaled_x, scaled_y = frustum_scale(self.pos.y, 5*32, 5*32, w, h)
            scaled = pygame.transform.scale(self.explosion_frames[((pygame.time.get_ticks()-self.explosion_counter) // 10) % len(self.explosion_frames)], (int(scaled_x), int(scaled_y)))
            screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

        if self.stunned:
            x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z-self.h, w, h)
            scaled_x, scaled_y = frustum_scale(self.pos.y, 32, 32, w, h)
            scaled = pygame.transform.scale(self.stun_frames[(pygame.time.get_ticks() // 100) % len(self.stun_frames)], (int(scaled_x), int(scaled_y)))
            screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

        if self.confused:
            x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.z-self.h, w, h)
            scaled_x, scaled_y = frustum_scale(self.pos.y, 32, 32, w, h)
            scaled = pygame.transform.scale(self.confuse_frames[(pygame.time.get_ticks() // 100) % len(self.confuse_frames)], (int(scaled_x), int(scaled_y)))
            screen.blit(scaled, (x-scaled_x/2, y-scaled_y, scaled_x, scaled_y))

    def on_key_down(self, key):
        if key == self.button_left:
            self.left_pressed = True
        if key == self.button_right:
            self.right_pressed = True
        if key == self.button_up:
            self.up_pressed = True
        if key == self.button_down:
            self.down_pressed = True
        if key == self.button_jump:
            if not self.stunned:
                if not self.on_ground and self.vel_z > -20:
                    self.overheat += 60
                    if self.overheat > 100:
                        self.stunned = True
                        self.stun_start = pygame.time.get_ticks()
                    else:
                        self.pounding = True
                        self.vel_z = -20
                elif self.on_ground:
                    # self.charge_start = pygame.time.get_ticks()
                    # self.charging = True
                    self.vel_z = 10
                    self.jump_sound.play()

    def on_key_up(self, key):
        if key == self.button_left:
            self.left_pressed = False
        if key == self.button_right:
            self.right_pressed = False
        if key == self.button_up:
            self.up_pressed = False
        if key == self.button_down:
            self.down_pressed = False
        # if key == self.button_jump:
        #     if self.on_ground and self.charging:
        #         charge = min(1, (pygame.time.get_ticks() - self.charge_start) / 500)
        #         self.vel_z = 15*charge
        #         self.charging = False
        #         self.jump_sound.play()

class Particle:
    def __init__(self, lifetime, pos, vel, acc):
        self.lifetime = lifetime
        self.pos = pos
        self.vel = vel
        self.acc = acc
        self.alive = True

    def think(self, game, w, h):
        self.lifetime -= 1
        if self.pos.z < 0 or self.lifetime <= 0:
            self.alive = False
            return
        self.pos += self.vel
        self.vel += self.acc

    def draw(self, screen, w, h):
        if not self.alive:
            return
        x, y = frustum(self.pos.x+w/2, self.pos.y+h/2, -self.pos.z, w, h)
        pydraw.circle(screen, (255, 255, 255), (int(x), int(y)), 1)

class Game:
    listeners = {}
    objects = []
    waves = []
    particles = []
    character_datas = [
        CharacterData(
            ("new_anvil_walk_down_map.png"  , 11, 28, 26),
            ("new_anvil_walk_left_map.png"  , 14, 17, 26),
            ("new_anvil_walk_right_map.png" , 14, 17, 26),
            ("new_anvil_walk_up_map.png"    , 11, 28, 26),
        ),
        CharacterData(
            ("cropped_ukko_paikallaan.png" , 4, 20, 22),
            ("cropped_ukko_vasemmalle.png" , 4, 20, 22),
            ("cropped_ukko_oikealle.png"   , 4, 20, 22),
            ("cropped_ukko_taakse.png"     , 4, 20, 22),
        ),
        CharacterData(
            ("kaarme1etu.png"   , 4, 15, 17),
            ("kaarme1vasen.png" , 4, 19, 17),
            ("kaarme1oikea.png" , 4, 19, 17),
            ("kaarme1taka.png"  , 4, 15, 17),
        ),
        CharacterData(
            ("robotti1etu.png"  , 2, 17, 28),
            ("robotti1etu.png"  , 2, 17, 28),
            ("robotti1etu.png"  , 2, 17, 28),
            ("robotti1taka.png" , 2, 17, 28),
        ),
    ]

    menu_options = [
        ""
    ]

    def reset(self):
        randoms = random.sample(self.character_datas, 2)

        self.winner = None
        self.scores = [0, 0]

        self.pulu_x = 0

        self.listeners[EventType.KEYDOWN] = []
        self.listeners[EventType.KEYUP] = []
        self.objects = []
        player1 = Player(K_w, K_a, K_s, K_d, K_LCTRL, self, randoms[0])
        self.listen_event(EventType.KEYDOWN, player1.on_key_down)
        self.listen_event(EventType.KEYUP, player1.on_key_up)
        player1.pos.x = -640
        player1.pos.y = 0
        self.objects.append(player1)
        self.player1 = player1
        player2 = Player(K_i, K_j, K_k, K_l, K_RCTRL, self, randoms[1])
        self.listen_event(EventType.KEYDOWN, player2.on_key_down)
        self.listen_event(EventType.KEYUP, player2.on_key_up)
        player2.pos.x = 640
        player2.pos.y = 0
        self.objects.append(player2)
        self.player2 = player2
        ball = Ball()
        self.objects.append(ball)

        spring = Spring(v2(-630, -350), v2(40, 0))
        self.objects.append(spring)
        spring = Spring(v2(630, 350), v2(-40, 0))
        self.objects.append(spring)
        spring = Spring(v2(-630, 350), v2(40, 0))
        self.objects.append(spring)
        spring = Spring(v2(630, -350), v2(-40, 0))
        self.objects.append(spring)

        self.in_menu = True

    def start(self, target_score):
        self.target_score = target_score
        self.score_font = pygame.font.SysFont("monospace", 72)
        self.winner_font = pygame.font.Font("njnaruto.ttf", 48)
        self.hint_font = pygame.font.Font("njnaruto.ttf", 24)
        self.anime_font = pygame.font.Font("njnaruto.ttf", 36)

        self.pulu_images = []
        img = pygame.image.load("pulu.png")
        for i in range(5):
            sub = img.subsurface((0, i*20, 20, 20))
            sub = pygame.transform.scale(sub, (20, 20))
            self.pulu_images.append(sub)

        self.field = pygame.image.load("kentta3dv10fp.png")
        self.field = pygame.transform.scale(self.field, (1280, 720))

        self.title_image = pygame.image.load("title_image.png")

        self.canvas = pygame.display.set_mode((1280, 720))
        self.reset()
        self.run()

    def run(self):
        self.running = True
        self.clock = pygame.time.Clock()
        while self.running:
            info = self.canvas.get_size()
            w = info[0]
            h = info[1]
            self.clock.tick(60)
            if self.in_menu:
                self.menu_think(w, h)
                self.menu_draw(w, h)
            else:
                self.think(w, h)
                self.draw(w, h)

    def menu_think(self, w, h):
        for event in pygame.event.get():
            if event.type == QUIT:
                self.running = False
            elif event.type == KEYDOWN:
                if event.key == K_SPACE:
                    self.in_menu = False

    def menu_draw(self, w, h):
        self.canvas.blit(self.title_image, (0, 0))
        label = self.anime_font.render("Press spacebar to play!", 1, (255, 255, 255))
        scale = 1.2+math.sin(pygame.time.get_ticks()/500)*0.1
        label = pygame.transform.scale(label, (int(label.get_width()*scale), int(label.get_height()*scale)))
        self.canvas.blit(label, (w/2-label.get_width()/2, h-label.get_height()-100))
        pygame.display.update()

    def think(self, w, h):
        for event in pygame.event.get():
            if event.type == QUIT:
                self.running = False
            elif event.type == KEYDOWN:
                if self.winner and event.key == K_SPACE:
                    self.reset()
                    return
                for cb in self.listeners[EventType.KEYDOWN]:
                    cb(event.key)
            elif event.type == KEYUP:
                for cb in self.listeners[EventType.KEYUP]:
                    cb(event.key)

        if not self.winner:
            mmm = random.random()
            if mmm < 0.001:
                self.objects.append(MeteorPowerup(v2(random.randint(-640, 640), random.randint(-360, 360))))
            elif mmm < 0.004:
                power_type = random.choice([FreezePowerup, SwapPowerup, ThrowPowerup, ConfusePowerup])
                self.objects.append(power_type(v2(random.randint(-640, 640), random.randint(-360, 360))))

        self.pulu_x -= 5
        self.pulu_x = (self.pulu_x + 1280) % 1280

        if not self.winner:
            for wave in self.waves:
                wave.think(self, w, h)

            for obj in self.objects:
                obj.think(self, w, h)

            for par in self.particles:
                par.think(self, w, h)

            self.particles = [par for par in self.particles if par.alive]
            self.objects = [obj for obj in self.objects if obj.alive]
            self.waves = [wave for wave in self.waves if wave.alive]

            # If we've reached an end condition...
            if self.scores[0] == self.target_score:
                self.winner = "Player 1"
            elif self.scores[1] == self.target_score:
                self.winner = "Player 2"

    def draw(self, w, h):
        self.canvas.fill((0, 255, 0))
        self.canvas.blit(self.field, (0, 0, w, h))
        # corners = [
        #     (frustum(0, 0, 0, w, h)),
        #     (frustum(0, h, 0, w, h)),
        #     (frustum(w, h, 0, w, h)),
        #     (frustum(w, 0, 0, w, h))
        # ]
        # pydraw.lines(self.canvas, (255, 255, 255), True, corners)
        # pydraw.line(self.canvas, (255, 0, 0), (frustum(0, h/2, 0, w, h)), (frustum(w, h/2, 0, w, h)))
        # pydraw.line(self.canvas, (255, 0, 0), (frustum(0, h/2-GOAL_RANGE*h, 0, w, h)), (frustum(w, h/2-GOAL_RANGE*h, 0, w, h)))
        # pydraw.line(self.canvas, (255, 0, 0), (frustum(0, h/2+GOAL_RANGE*h, 0, w, h)), (frustum(w, h/2+GOAL_RANGE*h, 0, w, h)))

        label = self.score_font.render("{}".format(self.scores[0]), 1, (255, 255, 255))
        self.canvas.blit(label, (0, h-label.get_height()))

        label = self.score_font.render("{}".format(self.scores[1]), 1, (255, 255, 255))
        self.canvas.blit(label, (w-label.get_width(), h-label.get_height()))

        pydraw.rect(self.canvas, (0, 0, 0), (0, h-15, 200, 10))
        pydraw.rect(self.canvas, (255, 255, 255), (0, h-15, 200*(min(1, self.player1.overheat/100)), 10))

        pydraw.rect(self.canvas, (0, 0, 0), (w-200, h-15, 200, 10))
        sw = 200*(min(1, self.player2.overheat/100))
        pydraw.rect(self.canvas, (255, 255, 255), (w-sw, h-15, sw, 10))

        self.objects.sort(key=lambda o: o.pos.y)

        for wave in self.waves:
            wave.draw(self.canvas, w, h)

        for obj in self.objects:
            obj.draw(self.canvas, w, h)

        for par in self.particles:
            par.draw(self.canvas, w, h)

        self.canvas.blit(self.pulu_images[(pygame.time.get_ticks() // 100) % len(self.pulu_images)], (self.pulu_x, 20 + math.sin(pygame.time.get_ticks() / 500) * 20))

        if self.winner:
            label = self.winner_font.render("{} wins!".format(self.winner), 1, (255, 0, 0))
            self.canvas.blit(label, (w/2-label.get_width()/2, h/2-label.get_height()/2))
            label = self.hint_font.render("Hit spacebar to restart", 1, (255, 0, 0))
            self.canvas.blit(label, (w/2-label.get_width()/2, h/2-label.get_height()/2 + 80))

        pygame.display.update()

    def listen_event(self, type, callback):
        self.listeners[type].append(callback)

    def spawn_particle(self, particle):
        self.particles.append(particle)

def main():
    pygame.init()
    pygame.display.set_caption("Power Stomp Ball - Made with 💟 at FGJ")
    game = Game()
    game.start(3)
    pygame.quit()

if __name__ == "__main__":
    main()
